@@ -21,6 +21,8 @@ except ImportError as e:
     sys.stderr.write(f"ERROR : {e} : install it to use this utility...")
     sys.exit(1)
 
+# copy rather than link for any datasets with product of extents less than the following
+MIN_DSET_ELEMENTS_FOR_LINKING=512
 
 def dump_dtype(dt):
     if not isinstance(dt, np.dtype):
@@ -72,7 +74,7 @@ def is_regionreference(val):
         elif isinstance(val, type) and val.__name__ == "RegionReference":
             return True
     except AttributeError as ae:
-        msg = f"is_reference for {val} error: {ae}"
+        msg = f"is_regionreference for {val} error: {ae}"
         logging.warning(msg)
 
     return False
@@ -96,6 +98,13 @@ def has_reference(dtype):
         has_ref = has_reference(basedt)
     return has_ref
 
+def get_reftype(obj):
+    if is_h5py(obj):
+        ref_type = h5py. special_dtype(ref=h5py.Reference)
+    else:
+        ref_type = h5pyd.special_dtype(ref=h5pyd.Reference)
+    return ref_type
+
 
 def is_vlen(dtype):
     if dtype.metadata and "vlen" in dtype.metadata:
@@ -305,7 +314,7 @@ def copy_array(src_arr, ctx):
 def copy_attribute(desobj, name, srcobj, ctx):
 
     msg = f"creating attribute {name} in {srcobj.name}"
-    logging.debug(msg)
+    logging.info(msg)
 
     if ctx["verbose"]:
         print(msg)
@@ -313,10 +322,51 @@ def copy_attribute(desobj, name, srcobj, ctx):
     tgtarr = None
     data = srcobj.attrs[name]
     src_dt = None
+    
+    # check for non-numpy types that might get returned
+    if is_regionreference(data):
+        msg = "regionreference types not supported, "
+        msg += f"attribute {name} in object {desobj.name} will not be loaded"
+        if ctx["verbose"]:
+            print(msg)
+        logging.warning(msg)
+        return
+
+    if is_reference(data):
+        src_dt = get_reftype(srcobj)
+        tgt_dt = get_reftype(desobj)
+        tgt_ref = copy_element(data, src_dt, tgt_dt, ctx)
+        try:
+            desobj.attrs.create(name, tgt_ref)
+        except (IOError, TypeError) as e:
+            msg = f"ERROR: failed to create attribute {name} "
+            msg += f"of object {desobj.name} for reference type -- {e}"
+            logging.error(msg)
+            if not ctx["ignore_error"]:
+                raise IOError(msg)
+
+        # done with non-numpy compatible data
+        return
+
     try:
         src_dt = data.dtype
     except AttributeError:
-        pass  # auto convert to numpy array
+        # convert to numpy type
+        data = np.asarray(data)
+        src_dt = data.dtype
+        
+    if src_dt.kind == "S" and isinstance(data, bytes):
+        # check that this is actually utf-encodable
+        try:
+            data.decode("utf-8")
+        except UnicodeDecodeError:
+            msg = f"byte value for attribute {name} in {srcobj.name} "
+            msg += "is not utf8 encodable - using surrogateescaping"
+            logging.warning(msg)
+            if ctx["verbose"]:
+                print(msg)
+            data = data.decode("utf-8", errors="surrogateescape")
+
     # First, make sure we have a NumPy array.
     if is_h5py(srcobj):
         src_empty = h5py.Empty
@@ -327,6 +377,7 @@ def copy_attribute(desobj, name, srcobj, ctx):
     else:
         des_empty = h5pyd.Empty
 
+    
     if isinstance(data, src_empty):
         # create Empty object with tgt dtype
         tgt_dt = convert_dtype(src_dt, ctx)
@@ -334,11 +385,12 @@ def copy_attribute(desobj, name, srcobj, ctx):
     else:
         srcarr = np.asarray(data, order="C", dtype=src_dt)
         tgtarr = copy_array(srcarr, ctx)
+
     try:
         desobj.attrs.create(name, tgtarr)
     except (IOError, TypeError) as e:
         msg = f"ERROR: failed to create attribute {name} "
-        msg += f"of object {desobj.naame} -- {e}"
+        msg += f"of object {desobj.name} -- {e}"
         logging.error(msg)
         if not ctx["ignore_error"]:
             raise IOError(msg)
@@ -413,11 +465,25 @@ def get_chunktable_dims(dset):
 
 # ----------------------------------------------------------------------------------
 def get_num_chunks(dset):
-    if dset.chunks:
+    if dset.shape is None:
+        # null space dataset - no data
+        return 0
+    elif np.prod(dset.shape) == 0:
+        # zero extent dataset
+        return 0
+    elif np.prod(dset.shape) == 1:
+        # scalar or extent 1 dataset
+        return 1
+    elif dset.chunks:
         if is_h5py(dset):
             dsetid = dset.id
             spaceid = dsetid.get_space()
-            num_chunks = dsetid.get_num_chunks(spaceid)
+            try:
+                num_chunks = dsetid.get_num_chunks(spaceid)
+            except ValueError:
+                # thrown for compact or contiguous datasets,
+                # just treat as 1 chunk
+                return 1
         else:
             # for hsds, just return maximum number of chunk in dataset
             chunk_table_dims = get_chunktable_dims(dset)
@@ -427,6 +493,18 @@ def get_num_chunks(dset):
         num_chunks = 1
     return num_chunks
 
+# ----------------------------------------------------------------------------------
+def get_dset_offset(dset):
+    """ Return dataset file offset for HDF5 contiguous datasets """
+    if not is_h5py(dset):
+        return -1
+    if dset.chunks is not None:
+        return -1
+    offset = dset.id.get_offset()
+    if offset is None:
+        return -1
+    return offset
+   
 # ----------------------------------------------------------------------------------
 def get_chunktable_dtype(include_file_uri=False):
     if include_file_uri:
@@ -492,7 +570,7 @@ def create_chunktable(dset, dset_dims, ctx):
         # use contiguous mapping
         chunks["class"] = "H5D_CONTIGUOUS_REF"
         chunks["file_uri"] = ctx["s3path"]
-        dset_offset = dset.id.get_offset()
+        dset_offset = get_dset_offset(dset)
         if dset_offset <= 0:
             msg = f"unexpected dataset_offset: {dset_offset}"
             logging.error(msg)
@@ -599,7 +677,14 @@ def update_chunktable(src, tgt, ctx):
    
     if is_h5py(src):   
         if src.chunks is None:
-            chunk_offset = src.id.get_offset()
+            chunk_offset = get_dset_offset(src)
+            if chunk_offset <= 0:
+                msg = "Expected dset_offset to be greater than zero"
+                logging.error(msg)
+                if not ctx["ignore_error"]:
+                    raise IOError(msg)
+                return
+
             chunk_size = src.id.get_storage_size()
             if extend:
                 chunkinfo_arr[...] = (chunk_offset, chunk_size, s3path)
@@ -694,7 +779,7 @@ def create_dataset(dobj, ctx):
     dset = None
     dset_preappend = None
 
-    msg = f"creating dataset {dobj.name}, shape: {dobj.shape}, type: {dobj.dtype}"
+    msg = f"create_dataset({dobj.name})"
     logging.info(msg)
     if ctx["verbose"]:
         print(msg)
@@ -703,6 +788,12 @@ def create_dataset(dobj, ctx):
     if dobj.name in fout:
         dset = fout[dobj.name]
         logging.debug(f"{dobj.name} already exists")
+        if ctx["no_clobber"]:
+            msg = f"skipping creation of dataset {dobj.name} since already exist and no-clobber option is used"
+            logging.info(msg)
+            if ctx["verbose"]:
+                print(msg)
+            return
         if ctx["append"]:
             if ctx["extend_dim"]:
                 msg = f"skipping creation of dataset {dobj.name} since already found"
@@ -812,6 +903,8 @@ def create_dataset(dobj, ctx):
             and not is_vlen(dobj.dtype)
             and dobj.shape is not None
             and len(dobj.shape) > 0
+            and not is_compact(dobj)
+            and np.prod(dobj.shape) > MIN_DSET_ELEMENTS_FOR_LINKING
         ):
             chunks = create_chunktable(dobj, tgt_shape, ctx)
             logging.info(f"using chunk layout: {chunks}")
@@ -860,7 +953,11 @@ def create_dataset(dobj, ctx):
             logging.debug(f"got fillvalue: {fillvalue}")
             kwargs["fillvalue"] = fillvalue
 
-        # finally, create the dataset   
+        # finally, create the dataset  
+        msg = f"creating dataset {dobj.name}, shape: {dobj.shape}, type: {dobj.dtype}"
+        logging.info(msg)
+        if ctx["verbose"]:
+            print(msg) 
         dset = fout.create_dataset(dobj.name, **kwargs)
 
         msg = f"dataset created, uuid: {dset.id.id}, "
@@ -968,9 +1065,11 @@ def write_dataset(src, tgt, ctx):
             print(msg)
         resize_dataset(tgt, new_extent, axis=0)
 
-    if ctx["dataload"] == "link":
-        # don't write chunks, but update chunktable for chunk ref indirect
-        if tgt.id.layout and tgt.id.layout["class"] == "H5D_CHUNKED_REF_INDIRECT":
+
+    if tgt.id.layout and tgt.id.layout["class"] != "H5D_CHUNKED":
+        # this is one of the ref layouts
+        if tgt.id.layout["class"] == "H5D_CHUNKED_REF_INDIRECT":
+            # don't write chunks, but update chunktable for chunk ref indirect
             update_chunktable(src, tgt, ctx)
         else:
             pass # skip chunkterator for link option
@@ -1002,7 +1101,7 @@ def write_dataset(src, tgt, ctx):
 
         for src_s in it:
             logging.debug(f"src selection: {src_s}")
-            if rank == 1:
+            if rank == 1 and isinstance(src_s, slice):
                 start = src_s.start + offset[0]
                 stop = src_s.stop + offset[0]
                 if len(tgt.shape) > rank:
@@ -1134,6 +1233,13 @@ def create_group(gobj, ctx):
     if gobj.name in fout:
         grp = fout[gobj.name]
         logging.debug(f"{gobj.name} already exists")
+        if ctx["no_clobber"]:
+            msg = f"skipping creation of group {gobj.name} since already exists and no-clobber mode is used"
+            logging.info(msg)
+            if ctx["verbose"]:
+                print(msg)
+            return
+        
         if ctx["append"]:
             msg = f"skipping creation of group {gobj.name} since already found"
             logging.info(msg)
@@ -1186,6 +1292,7 @@ def load_file(
     compression=None,
     compression_opts=None,
     append=False,
+    no_clobber=False,
     extend_dim=None,
     extend_offset=0,
     ignore_error=False,
@@ -1217,6 +1324,7 @@ def load_file(
     ctx["default_compression_opts"] = compression_opts
     ctx["s3path"] = s3path
     ctx["append"] = append
+    ctx["no_clobber"] = no_clobber
     ctx["extend_dim"] = extend_dim
     ctx["extend_offset"] = extend_offset
     ctx["srcid_desobj_map"] = {}
